@@ -21,9 +21,10 @@ use std::str::from_utf8;
 use memchr::{memchr, memchr2};
 use nom::branch::alt;
 use nom::bytes::complete::tag;
-use nom::combinator::{cut, eof, not, opt};
-use nom::sequence::{preceded, terminated, tuple};
-use nom::IResult;
+use nom::combinator::opt;
+use nom::error::{Error, ErrorKind};
+use nom::sequence::{terminated, tuple};
+use nom::{error_position, IResult};
 use proc_macro::TokenStream;
 use quote::quote;
 use syn::{Attribute, DeriveInput, Expr, ExprLit, ExprPath, Lit};
@@ -102,7 +103,9 @@ const _: &'static [u8] = ::core::include_bytes!({:?});"#,
     parse_file(f, &mut content);
     writeln!(content, "::core::fmt::Result::Ok(())\n}}\n}}").unwrap();
 
-    content.parse().unwrap()
+    content
+        .parse()
+        .expect("Could not parse generated code as Rust source.")
 }
 
 enum TemplateAttr {
@@ -111,12 +114,12 @@ enum TemplateAttr {
 
 fn parse_attributes(attrs: Vec<Attribute>) -> TemplateAttr {
     let mut arguments = None;
-    for Attribute {
-        path: syn::Path { segments, .. },
-        tokens,
-        ..
-    } in attrs.into_iter()
-    {
+    for attr in attrs.into_iter() {
+        let Attribute {
+            path: syn::Path { segments, .. },
+            tokens,
+            ..
+        } = attr;
         if segments.len() == 1 && segments.first().unwrap().ident == "template" {
             if arguments.is_some() {
                 panic!("Duplicated #[template(…)] attribute.");
@@ -171,7 +174,9 @@ fn parse_attributes(attrs: Vec<Attribute>) -> TemplateAttr {
         }
     };
 
-    match syn::parse2(arguments).unwrap() {
+    match syn::parse2(arguments)
+        .expect("In #[template(…)] you need to use comma separated key=\"value\" pairs.")
+    {
         Expr::Paren(p) => handle_arg(&*p.expr),
         Expr::Tuple(g) => {
             for e in g.elems {
@@ -195,8 +200,11 @@ fn parse_attributes(attrs: Vec<Attribute>) -> TemplateAttr {
 
 fn parse_file(mut input: impl Read, mut output: impl Write) {
     let mut buf = Vec::new();
-    input.read_to_end(&mut buf).unwrap();
-    for blocks in parse(&buf).into_iter() {
+    input
+        .read_to_end(&mut buf)
+        .expect("Could not read source file even after successfully opening it.");
+    let p = parse(&buf);
+    for blocks in p.into_iter() {
         match blocks {
             CodeOrData::Code(blocks) => {
                 for code in blocks.into_iter() {
@@ -222,21 +230,37 @@ fn parse_file(mut input: impl Read, mut output: impl Write) {
                 for data in blocks.into_iter() {
                     match data {
                         DataSection::Data(_) => {}
-                        DataSection::Raw(s) => {
-                            writeln!(output, "    &({}),", s).unwrap();
-                        }
                         DataSection::Escaped(s) => {
-                            writeln!(output, "    ::nate::XmlEscape(&({})),", s).unwrap();
+                            writeln!(output, "::nate::XmlEscape(&({})),", s).unwrap();
                         }
-                        DataSection::Debug(s) => {
-                            writeln!(output, "    &({}),", s).unwrap();
-                        }
-                        DataSection::Verbose(s) => {
-                            writeln!(output, "    &({}),", s).unwrap();
+                        DataSection::Raw(s) | DataSection::Debug(s) | DataSection::Verbose(s) => {
+                            writeln!(output, "&({}),", s).unwrap();
                         }
                     }
                 }
-                writeln!(output, "\n)?;").unwrap();
+                writeln!(output, ")?;").unwrap();
+            }
+            CodeOrData::Error(nom::Err::Incomplete(_)) => {
+                panic!("Impossible");
+            }
+            CodeOrData::Error(nom::Err::Error(err) | nom::Err::Failure(err)) => {
+                let offset = buf.len() - err.input.len();
+                let (source_before, source_after) = from_utf8(&buf).unwrap().split_at(offset);
+
+                let source_after = match source_after.char_indices().enumerate().take(41).last() {
+                    Some((40, (i, _))) => format!("{:?}...", &source_after[..i]),
+                    _ => format!("{:?}", source_after),
+                };
+
+                let (row, last_line) = source_before.lines().enumerate().last().unwrap();
+                let column = last_line.chars().count();
+
+                panic!(
+                    "Problems parsing template source at row {}, column {} near:\n{}",
+                    row + 1,
+                    column,
+                    source_after,
+                );
             }
         }
     }
@@ -268,17 +292,20 @@ fn parse(i: &[u8]) -> Vec<CodeOrData<'_>> {
                     Some(CodeOrData::Data(blocks)) => blocks.push(data),
                     _ => accu.push(CodeOrData::Data(vec![data])),
                 },
+                Block::Error(err) => accu.push(CodeOrData::Error(err)),
             }
             accu
         })
 }
 
+#[derive(Debug)]
 enum CodeOrData<'a> {
     Code(Vec<&'a str>),
     Data(Vec<DataSection<'a>>),
+    Error(nom::Err<Error<&'a [u8]>>),
 }
 
-#[derive(Clone)]
+#[derive(Debug, Copy, Clone)]
 enum DataSection<'a> {
     Data(&'a str),
     Raw(&'a str),
@@ -287,11 +314,35 @@ enum DataSection<'a> {
     Verbose(&'a str),
 }
 
-#[derive(Clone)]
+#[derive(Debug)]
 enum Block<'a> {
     Data(DataSection<'a>),
     Code(&'a str),
     Comment,
+    Error(nom::Err<Error<&'a [u8]>>),
+}
+
+struct WsBlock<'a>(bool, Block<'a>, bool);
+
+struct WsBlockIter<'a>(&'a [u8], Option<WsBlock<'a>>);
+
+fn clone_err<'a>(err: &nom::Err<Error<&'a [u8]>>) -> nom::Err<Error<&'a [u8]>> {
+    match *err {
+        nom::Err::Incomplete(err) => nom::Err::Incomplete(err),
+        nom::Err::Error(Error { input, code }) => nom::Err::Error(Error { input, code }),
+        nom::Err::Failure(Error { input, code }) => nom::Err::Failure(Error { input, code }),
+    }
+}
+
+impl Clone for Block<'_> {
+    fn clone(&self) -> Self {
+        match *self {
+            Block::Data(d) => Block::Data(d),
+            Block::Code(c) => Block::Code(c),
+            Block::Comment => Block::Comment,
+            Block::Error(ref err) => Block::Error(clone_err(err)),
+        }
+    }
 }
 
 impl Block<'_> {
@@ -299,22 +350,20 @@ impl Block<'_> {
         match self {
             Block::Comment => true,
             Block::Code(s) | Block::Data(DataSection::Data(s)) => s.is_empty(),
-            Block::Data(_) => false,
+            Block::Data(_) | Block::Error(_) => false,
         }
     }
 }
 
-struct WsBlock<'a>(bool, Block<'a>, bool);
-
-struct WsBlockIter<'a>(&'a [u8], Option<WsBlock<'a>>);
-
 impl<'a> WsBlockIter<'a> {
     fn new(i: &'a [u8]) -> Self {
         if i.is_empty() {
             Self(b"", None)
         } else {
-            let (i, b) = parse_ws_block(i).unwrap();
-            Self(i, Some(b))
+            match parse_ws_block(i) {
+                Ok((i, b)) => Self(i, Some(b)),
+                Err(err) => Self(b"", Some(WsBlock(false, Block::Error(err), false))),
+            }
         }
     }
 }
@@ -324,9 +373,16 @@ impl<'a> Iterator for WsBlockIter<'a> {
 
     fn next(&mut self) -> Option<Self::Item> {
         let cur = if !self.0.is_empty() {
-            let (j, b) = parse_ws_block(self.0).unwrap();
-            self.0 = j;
-            Some(b)
+            match parse_ws_block(self.0) {
+                Ok((j, b)) => {
+                    self.0 = j;
+                    Some(b)
+                }
+                Err(err) => {
+                    self.0 = b"";
+                    Some(WsBlock(false, Block::Error(err), false))
+                }
+            }
         } else {
             None
         };
@@ -344,261 +400,71 @@ impl<'a> Iterator for WsBlockIter<'a> {
 
 fn parse_ws_block(i: &[u8]) -> IResult<&[u8], WsBlock<'_>> {
     alt((
-        parse_ws_verbose,
-        parse_ws_debug,
-        parse_ws_raw,
-        parse_ws_escaped,
-        parse_ws_comment,
-        parse_ws_code,
-        parse_ws_data,
+        |i| {
+            let (i, (a, b, z)) = parse_block(i, b"{%", b"%}")?;
+            Ok((i, WsBlock(a, Block::Code(b), z)))
+        },
+        |i| {
+            let (i, (a, _, z)) = parse_block(i, b"{#", b"#}")?;
+            Ok((i, WsBlock(a, Block::Comment, z)))
+        },
+        |i| parse_data_section(i, b"{{{{{", b"}}}}}", DataSection::Verbose),
+        |i| parse_data_section(i, b"{{{{", b"}}}}", DataSection::Debug),
+        |i| parse_data_section(i, b"{{{", b"}}}", DataSection::Raw),
+        |i| parse_data_section(i, b"{{", b"}}", DataSection::Escaped),
+        parse_data,
     ))(i)
 }
 
-fn parse_ws_data(i: &[u8]) -> IResult<&[u8], WsBlock<'_>> {
-    let (i, b) = parse_data(i)?;
-    Ok((i, WsBlock(false, b, false)))
-}
-
-fn parse_ws_verbose(i: &[u8]) -> IResult<&[u8], WsBlock<'_>> {
-    let (i, (a, b, z)) = preceded(
-        tag(b"{{{{{"),
-        cut(terminated(
-            tuple((opt(tag(b"-")), parse_verbose, opt(tag(b"-")))),
-            tag(b"}}}}}"),
-        )),
-    )(i)?;
-    Ok((i, WsBlock(a.is_some(), b, z.is_some())))
-}
-
-fn parse_ws_debug(i: &[u8]) -> IResult<&[u8], WsBlock<'_>> {
-    let (i, (a, b, z)) = preceded(
-        tag(b"{{{{"),
-        cut(terminated(
-            tuple((opt(tag(b"-")), parse_debug, opt(tag(b"-")))),
-            tag(b"}}}}"),
-        )),
-    )(i)?;
-    Ok((i, WsBlock(a.is_some(), b, z.is_some())))
-}
-
-fn parse_ws_raw(i: &[u8]) -> IResult<&[u8], WsBlock<'_>> {
-    let (i, (a, b, z)) = preceded(
-        tag(b"{{{"),
-        cut(terminated(
-            tuple((opt(tag(b"-")), parse_raw, opt(tag(b"-")))),
-            tag(b"}}}"),
-        )),
-    )(i)?;
-    Ok((i, WsBlock(a.is_some(), b, z.is_some())))
-}
-
-fn parse_ws_escaped(i: &[u8]) -> IResult<&[u8], WsBlock<'_>> {
-    let (i, (a, b, z)) = preceded(
-        tag(b"{{"),
-        cut(terminated(
-            tuple((opt(tag(b"-")), parse_escaped, opt(tag(b"-")))),
-            tag(b"}}"),
-        )),
-    )(i)?;
-    Ok((i, WsBlock(a.is_some(), b, z.is_some())))
-}
-
-fn parse_ws_comment(i: &[u8]) -> IResult<&[u8], WsBlock<'_>> {
-    let (i, (a, b, z)) = preceded(
-        tag(b"{#"),
-        cut(tuple((
-            opt(tag(b"-")),
-            parse_comment,
-            opt(terminated(opt(tag(b"-")), tag(b"#}"))),
-        ))),
-    )(i)?;
-    Ok((i, WsBlock(a.is_some(), b, z.is_some())))
-}
-
-fn parse_ws_code(i: &[u8]) -> IResult<&[u8], WsBlock<'_>> {
-    let (i, (a, b, z)) = preceded(
-        tag(b"{%"),
-        cut(tuple((
-            opt(tag(b"-")),
-            parse_code,
-            opt(terminated(opt(tag(b"-")), tag(b"%}"))),
-        ))),
-    )(i)?;
-    Ok((i, WsBlock(a.is_some(), b, z.is_some())))
-}
-
-fn parse_comment(i: &[u8]) -> IResult<&[u8], Block<'_>> {
-    fn end(i: &[u8]) -> IResult<&[u8], &[u8]> {
-        alt((tag(b"-#}"), tag("#}")))(i)
-    }
-    if i.len() >= 2 {
+fn parse_block<'a>(
+    i: &'a [u8],
+    start: &'static [u8],
+    end: &'static [u8],
+) -> IResult<&'a [u8], (bool, &'a str, bool)> {
+    let inner = |i: &'a [u8]| -> IResult<&'a [u8], (&'a str, bool)> {
         let mut start = 0;
-        while start < i.len() {
-            match memchr2(b'#', b'-', &i[start..=i.len() - 2]) {
-                Some(pos) if end(&i[pos..]).is_ok() => {
-                    return Ok((&i[start + pos..], Block::Comment))
-                }
-                Some(pos) => {
+        while (i.len() - start) >= end.len() {
+            if let Some(pos) = memchr2(end[0], b'-', &i[start..]) {
+                let (j, end) = opt(terminated(opt(tag(b"-")), tag(end)))(&i[start + pos..])?;
+                if let Some(trim) = end {
+                    return Ok((j, (from_utf8(&i[..start + pos]).unwrap(), trim.is_some())));
+                } else if pos > 0 {
                     start += pos;
+                } else {
+                    start += 1;
                 }
-                None => {
-                    break;
-                }
+            } else {
+                break;
             }
         }
-    }
-    Ok((b"", Block::Comment))
+        Ok((b"", (from_utf8(i).unwrap(), false)))
+    };
+    let (i, (_, trim_start, (b, trim_end))) = tuple((tag(start), opt(tag(b"-")), inner))(i)?;
+    Ok((i, (trim_start.is_some(), b, trim_end)))
 }
 
-fn parse_verbose(i: &[u8]) -> IResult<&[u8], Block<'_>> {
-    fn end(i: &[u8]) -> IResult<&[u8], &[u8]> {
-        alt((tag(b"-}}}}}"), tag("}}}}}")))(i)
+fn parse_data_section<'a>(
+    i: &'a [u8],
+    start: &'static [u8],
+    end: &'static [u8],
+    kind: impl 'static + Fn(&'a str) -> DataSection<'a>,
+) -> IResult<&'a [u8], WsBlock<'a>> {
+    let (j, (trim_start, b, trim_end)) = parse_block(i, start, end)?;
+    let b = b.trim();
+    if !b.is_empty() {
+        Ok((j, WsBlock(trim_start, Block::Data(kind(b)), trim_end)))
+    } else {
+        Err(nom::Err::Error(error_position!(i, ErrorKind::NonEmpty)))
     }
-    if i.len() >= 5 {
-        let mut start = 0;
-        while start < i.len() {
-            match memchr2(b'}', b'-', &i[start..=i.len() - 5]) {
-                Some(pos) if end(&i[pos..]).is_ok() => {
-                    return Ok((
-                        &i[start + pos..],
-                        Block::Data(DataSection::Verbose(from_utf8(&i[..start + pos]).unwrap())),
-                    ))
-                }
-                Some(pos) => {
-                    start += pos;
-                }
-                None => {
-                    break;
-                }
-            }
-        }
-    }
-    Err(nom::Err::Error(nom::error::Error {
-        input: i,
-        code: nom::error::ErrorKind::Tag,
-    }))
 }
 
-fn parse_debug(i: &[u8]) -> IResult<&[u8], Block<'_>> {
-    fn end(i: &[u8]) -> IResult<&[u8], &[u8]> {
-        alt((tag(b"-}}}}"), tag("}}}}")))(i)
-    }
-    if i.len() >= 4 {
-        let mut start = 0;
-        while start < i.len() {
-            match memchr2(b'}', b'-', &i[start..=i.len() - 4]) {
-                Some(pos) if end(&i[pos..]).is_ok() => {
-                    return Ok((
-                        &i[start + pos..],
-                        Block::Data(DataSection::Debug(from_utf8(&i[..start + pos]).unwrap())),
-                    ))
-                }
-                Some(pos) => {
-                    start += pos;
-                }
-                None => {
-                    break;
-                }
-            }
-        }
-    }
-    Err(nom::Err::Error(nom::error::Error {
-        input: i,
-        code: nom::error::ErrorKind::Tag,
-    }))
-}
-
-fn parse_raw(i: &[u8]) -> IResult<&[u8], Block<'_>> {
-    fn end(i: &[u8]) -> IResult<&[u8], &[u8]> {
-        alt((tag(b"-}}}"), tag("}}}")))(i)
-    }
-    if i.len() >= 3 {
-        let mut start = 0;
-        while start < i.len() {
-            match memchr2(b'}', b'-', &i[start..=i.len() - 3]) {
-                Some(pos) if end(&i[pos..]).is_ok() => {
-                    return Ok((
-                        &i[start + pos..],
-                        Block::Data(DataSection::Raw(from_utf8(&i[..start + pos]).unwrap())),
-                    ))
-                }
-                Some(pos) => {
-                    start += pos;
-                }
-                None => {
-                    break;
-                }
-            }
-        }
-    }
-    Err(nom::Err::Error(nom::error::Error {
-        input: i,
-        code: nom::error::ErrorKind::Tag,
-    }))
-}
-
-fn parse_escaped(i: &[u8]) -> IResult<&[u8], Block<'_>> {
-    fn end(i: &[u8]) -> IResult<&[u8], &[u8]> {
-        alt((tag(b"-}}"), tag("}}")))(i)
-    }
-    if i.len() >= 2 {
-        let mut start = 0;
-        while start < i.len() {
-            match memchr2(b'}', b'-', &i[start..=i.len() - 2]) {
-                Some(pos) if end(&i[pos..]).is_ok() => {
-                    return Ok((
-                        &i[start + pos..],
-                        Block::Data(DataSection::Escaped(from_utf8(&i[..start + pos]).unwrap())),
-                    ))
-                }
-                Some(pos) => {
-                    start += pos;
-                }
-                None => {
-                    break;
-                }
-            }
-        }
-    }
-    Err(nom::Err::Error(nom::error::Error {
-        input: i,
-        code: nom::error::ErrorKind::Tag,
-    }))
-}
-
-fn parse_code(i: &[u8]) -> IResult<&[u8], Block<'_>> {
-    fn end(i: &[u8]) -> IResult<&[u8], &[u8]> {
-        alt((tag(b"-%}"), tag("%}")))(i)
-    }
-    if i.len() >= 2 {
-        let mut start = 0;
-        while start < i.len() {
-            match memchr2(b'%', b'-', &i[start..=i.len() - 2]) {
-                Some(pos) if end(&i[pos..]).is_ok() => {
-                    return Ok((
-                        &i[start + pos..],
-                        Block::Code(from_utf8(&i[..start + pos]).unwrap()),
-                    ))
-                }
-                Some(pos) => {
-                    start += pos;
-                }
-                None => {
-                    break;
-                }
-            }
-        }
-    }
-    Ok((b"", Block::Code(from_utf8(i).unwrap())))
-}
-
-fn parse_data(i: &[u8]) -> IResult<&[u8], Block<'_>> {
-    let (i, _) = not(eof)(i)?;
-    match memchr(b'{', &i[1..]) {
-        Some(pos) => Ok((
-            &i[pos + 1..],
-            Block::Data(DataSection::Data(from_utf8(&i[..pos + 1]).unwrap())),
-        )),
-        None => Ok((b"", Block::Data(DataSection::Data(from_utf8(i).unwrap())))),
-    }
+fn parse_data<'a>(i: &'a [u8]) -> IResult<&'a [u8], WsBlock<'a>> {
+    let (i, b): (&'a [u8], _) = if let Some(pos) = memchr(b'{', &i[1..]) {
+        let (b, i) = i.split_at(pos + 1);
+        (i, b)
+    } else {
+        (b"", i)
+    };
+    let b = DataSection::Data(from_utf8(b).unwrap());
+    Ok((i, WsBlock(false, Block::Data(b), false)))
 }
